;                            Software License Agreement
;
; The software supplied herewith by Microchip Technology Incorporated (the "Company")
; for its PICmicro Microcontroller is intended and supplied to you, the Companys
; customer, for use solely and exclusively on Microchip PICmicro Microcontroller
; products.
;
; The software is owned by the Company and/or its supplier, and is protected under
; applicable copyright laws. All rights are reserved. Any use in violation of the
; foregoing restrictions may subject the user to criminal sanctions under applicable
; laws, as well as to civil liability for the breach of the terms and conditions of
; this license.
;
; THIS SOFTWARE IS PROVIDED IN AN "AS IS" CONDITION. NO WARRANTIES, WHETHER EXPRESS,
; IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
; MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. THE
; COMPANY SHALL NOT, IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL OR
; CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
;
; ###############################################################################
	TITLE "PIC with Flash EE data memory Interface"
;#define FOURMHZ
;#define TENMHZ
#define  TWENTYMHZ  
;
;       Program:          FLASH625.ASM
;       Revision Date:   
;               V1.00   30 June 1998      Adapted to 16CE62x parts
;		V1.10	19 February 2001  Several bug fixes and enhancements
;			* jump table may spill across page boundaries
;			* bank bits are saved and restored at the close of the function
;			* Delays encapsulated inside a macro to improve readability
;
; PIC16CE62x EEPROM communication code.  This code should be linked in
; with the application.  Uses 6 file registers in the common area (70-7F).
; These are only used during execution and may be overwritten between calls.
; No data is stored from one call to the next.  (No static or persistent data).
;
; If the operation is successful, bit 7 of PC_OFFSET will be set, and
; the functions will return W=1.  If the memory is busy with a write
; cycle, it will not ACK the command.  The functions will return with
; bit 7 of PC_OFFSET cleared and and W will be set to 0.
; 
; Based on Franco code.
;
; This provides users with highly compressed assembly code for
; communication between the EEPROM and the Microcontroller, which
; leaves a maximum amount of code space for the core application.
;
; Conditional assembly delays are included to meet standard mode timing
; specs.  For 4Mhz, define FOURMHZ at top of file.  For 10 Mhz, define TENMHZ.
;
; This code is specifically written for the interface hardware of the
; 16CE623/624/625 parts.  See AN571 for the unmodified routines.
;***************************************************************************
;***************************  EEPROM Subroutines  **************************
;***************************************************************************
; Communication for EEPROM based on I2C protocol, with Acknowledge.
;
; Byte_Write: Byte write routine
;       Inputs:  EEPROM Address   EEADDR
;                EEPROM Data      EEDATA
;       Outputs: Return 01 in W if OK, else return 00 in W
;
; Page_Write: Page write routine - writes up to 8 bytes at a time
;       Inputs:  FSR    points to beginning of RAM buffer.
;		 W      number of bytes to write
;		 EEADDR	EEPROM Address to start the write
;       Outputs: Return 01 in W if OK, else return 00 in W
;
; Read_Current: Read EEPROM at address currently held by EE device. 
;       Inputs:  NONE
;       Outputs: EEPROM Data       EEDATA
;                Return 01 in W if OK, else return 00 in W
;
; Read_Random: Read EEPROM byte at supplied address
;       Inputs:  EEPROM Address    EEADDR
;       Outputs: EEPROM Data       EEDATA
;                Return 01 in W if OK, else return 00 in W
;
; Page_Read : Page read routine - writes up to 8 bytes at a time
;       Inputs:  FSR    points to beginning of RAM buffer.
;		 W      number of bytes to read
;		 EEADDR EEPROM Address to start the read
;       Outputs: Return 01 in W if OK, else return 00 in W
;
; Note: EEPROM subroutines will set bit 7 in PC_OFFSET register if the
;       EEPROM acknowledged OK, else that bit will be cleared.  This bit 
;       can be checked instead of refering to the value returned in W
;
;***************************************************************************
;
; OPERATION:
;     Byte Write:
;       load EEADDR and EEDATA
;       then CALL WRITE_BYTE
;
;     Page Write:
;		 Load EEADDR
;		 Load FSR with address of 1st byte to transfer
;		 Load W with number of bytes to transfer (8 bytes max)
;		 then CALL WRITE_PAGE
;
;     Read Random:
;       Load EEADDR
;       then CALL READ_RANDOM
;       data read returned in EEDATA
;
;     Read Current
;       no setup necessary
;       CALL READ_CURRENT
;       data read returned in EEDATA
;
;***************************************************************************
;***************************  Variable Listing  ****************************
;***************************************************************************
OK          EQU     01H
NO          EQU     00H
EE_OK       EQU     07H         ; Bit 7 in PC_OFFSET used as OK flag for EE

; These must be in common RAM (addresses 70-7f).  This prevents excessive
; bank switching.
	 cblock	0x70
EEADDR			; EEPROM Address
EEDATA			; EEPROM Data
EEBYTE			; Byte sent to or received from
PC_OFFSET		; PC offset register (low order 4 bits), 
 			; EEPROM (control, address, or data)
bytecount		; # of bytes to write
			; Also, bit 7 used for EE_OK flag
Status_Temp		; Save the bank bits.
	endc	

DELAY4700NS		 MACRO
#ifdef  FOURMHZ
	call    delay4  ; wait 4 cycles  Tsu:sto (4.7 us)
#endif
#ifdef  TENMHZ
	call    delay10
#endif
#ifdef  TWENTYMHZ
	call    delay20
#endif
	ENDM

;********************** Set up EEPROM control bytes ************************
;***************************************************************************

READ_CURRENT
	movf	STATUS,w	; Save the bank bits
	movwf	Status_Temp
	movlw   B'10000100'	; PC offset for read current addr.  EE_OK bit7='1'
	movwf   PC_OFFSET	; Load PC offset
	bsf	STATUS,RP0	; set register page 1
	clrf	PCLATH		 		 
	goto    INIT_READ_CONTROL

WRITE_BYTE
	movf	STATUS,w	; Save the bank bits
	movwf	Status_Temp
	movlw   B'10000000'   ; PC offset for write byte.  EE_OK: bit7 = '1'
	goto    INIT_WRITE_CONTROL

WRITE_PAGE
	movwf	bytecount	; save off number of bytes to send
	movf	STATUS,w	; Save the bank bits
	movwf	Status_Temp
	movlw	B'10000111'   ; PC offset for write page.  EE_OK bit = 1
	goto	INIT_WRITE_CONTROL

READ_PAGE
	movwf	bytecount	; save off number of bytes to send
	movf	STATUS,w	; Save the bank bits
	movwf	Status_Temp
	movlw	B'10001010'	; PC offset for read page.  EE_OK bit = 1
	goto	INIT_WRITE_CONTROL

READ_RANDOM
	movf	STATUS,w
	movwf	Status_Temp
	movlw   B'10000011'	; PC offset for read random.  EE_OK: bit7 = '1'

INIT_WRITE_CONTROL
	movwf   PC_OFFSET	; Load PC offset register, value preset in W
	movlw   B'10100000'	; Control byte with write bit, bit 0 = '0'
	bsf	STATUS,RP0	; set register page 1
  
START_BIT
	clrf	PCLATH		 		 
	bcf     EEINTF,EESDA	; Start bit, EESDA and EESCL preset to '1'

;******* Set up output data (control, address, or data) and counter ********
;***************************************************************************
PREP_TRANSFER_BYTE
	movwf   EEBYTE		; Byte to transfer to EEPROM already in W
	bsf	STATUS,C	; set carry
	rlf	EEBYTE,F	; rotate bit into Carry

;************  Clock out data (control, address, or data) byte  ************
;***************************************************************************
OUTPUT_BYTE

	DELAY4700NS

	bcf     EEINTF,EESCL	; Set clock low during data set-up

	btfsc	STATUS,C		 
	bsf     EEINTF,EESDA	; Set data low, if rotated carry bit is
	btfss	STATUS,C
	bcf     EEINTF,EESDA	; reset data pin to a one, otherwise leave low
	bcf	STATUS,C	; clear carry

	DELAY4700NS
	bsf     EEINTF,EESCL	; clock data into EEPROM
	rlf	EEBYTE,F
	movf	EEBYTE,F	; test to see if it's zero
	btfss	STATUS,Z
	goto    OUTPUT_BYTE

	DELAY4700NS

;**************************  Acknowledge Check *****************************
;***************************************************************************

	bcf     EEINTF,EESCL	; Set EESCL low, 0.5us < ack valid < 3us

	DELAY4700NS

	bsf     EEINTF,EESDA    ; set data line high to check for acknowledge

	DELAY4700NS

	bsf     EEINTF,EESCL	; Raise EESCL, EEPROM acknowledge still valid

	DELAY4700NS

	btfsc   EEINTF,EESDA	; Check EESDA for acknowledge (low)
	bcf     PC_OFFSET,EE_OK ; If EESDA not low (no ack), set error flag

	DELAY4700NS

	bcf     EEINTF,EESCL    ; Lower EESCL, EEPROM release bus
	btfss   PC_OFFSET,EE_OK ; If no error continue, else stop bit
	goto    STOP_BIT


;*****  Set up program counter offset, based on EEPROM operating mode  *****
;***************************************************************************
STATEMACHINE
	movf    PC_OFFSET,W
	andlw   B'00001111'
	movwf	EEBYTE
	movlw	HIGH JUMPER
	movwf	PCLATH
	movlw	LOW JUMPER
	addwf   EEBYTE, W
	btfsc	STATUS,C
	incf	PCLATH,F
	movwf	PCL
JUMPER	goto    INIT_ADDRESS		; PC offset=0, write control done, send address
	goto    INIT_WRITE_DATA		; PC offset=1, write address done, send data
	goto    STOP_BIT		; PC offset=2, write done, send stop bit
	goto    INIT_ADDRESS		; PC offset=3, write control done, send address
	goto    INIT_READ_CONTROL	; PC offset=4, send read control
	goto    READ_BIT_COUNTER	; PC offset=5, set counter and read byte
	goto    STOP_BIT		; PC offset=6, random read done, send stop
	goto    INIT_ADDRESS		; PC offset=7, write control done, send address
	goto    INIT_WRITE_PAGE_DATA	; PC offset=8, write address done, send data
	goto    STOP_BIT		; PC offset=9, write done, send stop bit
	goto    INIT_ADDRESS		; PC offset=A, write control done, send address
	goto    INIT_READ_PAGE_CONTROL	; PC offset=B, write address done, send data
	goto    READ_PAGE_BIT_COUNTER	; PC offset=C, set counter and read byte


;**********  Initalize EEPROM data (address, data, or control) bytes  ******
;***************************************************************************
INIT_ADDRESS
	incf    PC_OFFSET, F	; Increment PC offset to 2 (write) or to 4 (read)
	movf    EEADDR,W	; Put EEPROM address in W, ready to send to EEPROM
	goto    PREP_TRANSFER_BYTE

INIT_WRITE_DATA
	incf    PC_OFFSET, F	; Increment PC offset to go to STOP_BIT next
	movf    EEDATA,W	; Put EEPROM data in W, ready to send to EEPROM
	goto    PREP_TRANSFER_BYTE

INIT_WRITE_PAGE_DATA
	decfsz	bytecount,f	; count byte tx'd
	goto	$+2		; If we're not done, don't increment to next state
	incf    PC_OFFSET, F    ; Increment PC offset to go to STOP_BIT next
	movf	INDF,W
	incf	FSR,F		 		 ; bump pointer
	goto    PREP_TRANSFER_BYTE

INIT_READ_CONTROL
	bsf     EEINTF,EESCL	; Raise EESCL
    DELAY4700NS
	bsf     EEINTF,EESDA	; raise EESDA
	incf    PC_OFFSET,F	; Increment PC offset to go to READ_BIT_COUNTER next
	movlw   B'10100001'	; Set up read control byte, ready to send to EEPROM
    DELAY4700NS
	goto    START_BIT	; bit 0 = '1' for read operation


INIT_READ_PAGE_CONTROL
	bsf     EEINTF,EESCL	; Raise EESCL
    DELAY4700NS
	bsf     EEINTF,EESDA	; raise EESDA
	incf    PC_OFFSET,F	; Increment PC offset to go to READ_BIT_COUNTER next
	movlw   B'10100001'	; Set up read control byte, ready to send to EEPROM
    DELAY4700NS
	goto    START_BIT	; bit 0 = '1' for read operation

;**************************  Read EEPROM data  *****************************
;***************************************************************************
READ_PAGE_BIT_COUNTER
	bsf     EEINTF,EESDA	; set data bit to 1 so we're not pulling bus down.
	movlw	0x01
	movwf	EEDATA

READ_BYTE_RPC

	DELAY4700NS

	bsf     EEINTF,EESCL	; Raise EESCL, EESDA valid.  EESDA still input from ack
	bsf	STATUS,C	; Assume bit to be read = 1

	DELAY4700NS

	btfss   EEINTF,EESDA	; Check if EESDA = 1
	bcf	STATUS,C	; if EESDA not = 1 then clear carry bit
	rlf     EEDATA, F	; rotate carry bit (=EESDA) into EEDATA;
	bcf     EEINTF,EESCL	; Lower EESCL
	DELAY4700NS
	bsf     EEINTF,EESDA	; reset EESDA
	btfss	STATUS,C	; we're done when the Carry gets set.
	goto    READ_BYTE_RPC   ; Read next bit if not finished reading byte

	movf	EEDATA,w
	movwf	INDF		; write data to buffer
	incf	FSR,f		; increment buffer pointer
	decfsz	bytecount,f
	goto	SEND_ACK
SEND_NAK
	bsf	EEINTF,EESDA	; Send an ACK (More reads to come)
	DELAY4700NS
	bsf     EEINTF,EESCL	; 
	DELAY4700NS
	bcf     EEINTF,EESCL
	goto	STOP_BIT	;
SEND_ACK
	bcf	EEINTF,EESDA	; Send an ACK (More reads to come)
	DELAY4700NS
	bsf     EEINTF,EESCL	; 
	DELAY4700NS
	bcf     EEINTF,EESCL
	goto	READ_PAGE_BIT_COUNTER		 
		 
; end read page bit control

READ_BIT_COUNTER
	bsf     EEINTF,EESDA	; set data bit to 1 so we're not pulling bus down.
	movlw	0x01		; this bit will be rotated though the byte.
	movwf	EEDATA		; we're done when it gets to the carry.

READ_BYTE_RBC
	DELAY4700NS

	bsf     EEINTF,EESCL	; Raise EESCL, EESDA valid.  EESDA still input from ack
	bsf	STATUS,C	; Assume bit to be read = 1

	DELAY4700NS

	btfss   EEINTF,EESDA	; Check if EESDA = 1
	bcf	STATUS,C	; if EESDA not = 1 then clear carry bit
	rlf     EEDATA,F	; rotate carry bit (=EESDA) into EEDATA;
	bcf     EEINTF,EESCL	; Lower EESCL
	DELAY4700NS
	bsf     EEINTF,EESDA	; reset EESDA
	btfss	STATUS,C	; we're done when the carry gets set.
	goto    READ_BYTE_RBC	; Read next bit if not finished reading byte

   DELAY4700NS
	bsf     EEINTF,EESCL
	DELAY4700NS
	bcf     EEINTF,EESCL
;******************  Generate a STOP bit and RETURN  ***********************
;***************************************************************************
STOP_BIT
	bcf     EEINTF,EESDA	; EESDA=0, on TRIS, to prepare for transition to '1' 
	bsf     EEINTF,EESCL	; EESCL = 1 to prepare for STOP bit

	DELAY4700NS

	bsf     EEINTF,EESDA	; Stop bit, EESDA transition to '1' while EESCL high
	bcf	STATUS,RP0
    
	movf	Status_Temp,w
	movwf	STATUS
	btfss   PC_OFFSET,EE_OK ; Check for error
	retlw   NO              ; if error, send back NO 
	retlw   OK		; if no error, send back OK

#ifdef TWENTYMHZ
delay20 goto    delay18
delay18 goto    delay16
delay16 goto    delay14
delay14 goto    delay12
delay12 goto    delay10
delay10 goto    delay8
delay8  goto    delay6
delay6  goto    delay4
delay4  return
#endif
#ifdef TENMHZ
; delay function.  Wait a number of cycles.  
delay10 goto    delay8
delay8  goto    delay6
delay6  goto    delay4
delay4  return
#endif
#ifdef FOURMHZ
; delay function.  Wait a number of cycles.  
delay4  return
#endif

;****************************************************************************
;************************  End EEPROM Subroutines  **************************

